@@ -3,110 +3,21 @@ The calculate module contains a routine for calculating the
 property surface of a system.
 """
 
-from __future__ import division
-from pycalphad import Model
-from pycalphad.model import DofError
-from pycalphad.core.sympydiff_utils import build_functions
+from pycalphad.codegen.callables import build_phase_records
+from pycalphad import ConditionError
 from pycalphad.core.utils import point_sample, generate_dof
 from pycalphad.core.utils import endmember_matrix, unpack_kwarg
-from pycalphad.core.utils import broadcast_to, unpack_condition, unpack_phases
+from pycalphad.core.utils import broadcast_to, filter_phases, unpack_condition,\
+    unpack_components, get_state_variables, instantiate_models
+from pycalphad.core.light_dataset import LightDataset
 from pycalphad.core.cache import cacheit
-from pycalphad.core.phase_rec import PhaseRecord, PhaseRecord_from_cython, PhaseRecord_from_compiledmodel
-from pycalphad.core.compiled_model import CompiledModel
+from pycalphad.core.phase_rec import PhaseRecord
 import pycalphad.variables as v
-from sympy import Symbol
 import numpy as np
 import itertools
 import collections
-import warnings
-from xarray import Dataset, Variable
-from xarray.core.dataset import as_dataset
-from xarray.core.combine import _calc_concat_over, _calc_concat_dim_coord, concat_vars
+from xarray import Dataset, concat
 from collections import OrderedDict
-import pandas as pd
-
-class FallbackModel(object):
-    "Compatibility layer while transitioning to CompiledModel."
-    def __new__(cls, *args, **kwargs):
-        try:
-            ret = CompiledModel(*args, **kwargs)
-        except NotImplementedError:
-            return Model(*args, **kwargs)
-        return ret
-
-def _fast_concat(datasets, dim=None):
-    """
-    Fork of xarray.concat for Datasets.
-    This function assumes everything is already aligned, and skips error checking.
-
-    Parameters
-    ----------
-    datasets
-
-    Returns
-    -------
-
-    """
-    data_vars = 'all'
-    coords = 'different'
-    positions = None
-
-    dim, coord = _calc_concat_dim_coord(dim)
-    datasets = [as_dataset(ds) for ds in datasets]
-
-    concat_over = _calc_concat_over(datasets, dim, data_vars, coords)
-
-    def insert_result_variable(k, v):
-        assert isinstance(v, Variable)
-        if k in datasets[0].coords:
-            result_coord_names.add(k)
-        result_vars[k] = v
-
-    # create the new dataset and add constant variables
-    result_vars = OrderedDict()
-    result_coord_names = set(datasets[0].coords)
-    result_attrs = datasets[0].attrs
-
-    for k, v in datasets[0].variables.items():
-        if k not in concat_over:
-            insert_result_variable(k, v)
-
-    # we've already verified everything is consistent; now, calculate
-    # shared dimension sizes so we can expand the necessary variables
-    dim_lengths = [ds.dims.get(dim, 1) for ds in datasets]
-    non_concat_dims = {}
-    for ds in datasets:
-        non_concat_dims.update(ds.dims)
-    non_concat_dims.pop(dim, None)
-
-    def ensure_common_dims(vars):
-        # ensure each variable with the given name shares the same
-        # dimensions and the same shape for all of them except along the
-        # concat dimension
-        common_dims = tuple(pd.unique([d for v in vars for d in v.dims]))
-        if dim not in common_dims:
-            common_dims = (dim,) + common_dims
-        for var, dim_len in zip(vars, dim_lengths):
-            if var.dims != common_dims:
-                common_shape = tuple(non_concat_dims.get(d, dim_len)
-                                     for d in common_dims)
-                var = var.expand_dims(common_dims, common_shape)
-            yield var
-
-    # stack up each variable to fill-out the dataset
-    for k in concat_over:
-        vars = ensure_common_dims([ds.variables[k] for ds in datasets])
-        combined = concat_vars(vars, dim, positions)
-        insert_result_variable(k, combined)
-
-    result = Dataset(result_vars, attrs=result_attrs)
-    result = result.set_coords(result_coord_names)
-
-    if coord is not None:
-        # add concat dimension last to ensure that its in the final Dataset
-        result[coord.name] = coord
-
-    return result
 
 
 def _generate_fake_points(components, statevar_dict, energy_limit, output, maximum_internal_dof, broadcast):
@@ -174,8 +85,9 @@ def _sample_phase_constitution(phase_name, phase_constituents, sublattice_dof, c
     vacancy_indices = list()
     for idx, sublattice in enumerate(phase_constituents):
         active_in_subl = sorted(set(phase_constituents[idx]).intersection(comps))
-        if 'VA' in active_in_subl and 'VA' in sorted(comps):
-            vacancy_indices.append(active_in_subl.index('VA'))
+        is_vacancy = [spec.number_of_atoms == 0 for spec in active_in_subl]
+        subl_va_indices = list(idx for idx, x in enumerate(is_vacancy) if x == True)
+        vacancy_indices.append(subl_va_indices)
     if len(vacancy_indices) != len(phase_constituents):
         vacancy_indices = None
     # Add all endmembers to guarantee their presence
@@ -197,24 +109,6 @@ def _sample_phase_constitution(phase_name, phase_constituents, sublattice_dof, c
                                  sampler(sublattice_dof,
                                          pdof=pdens)
                                  ))
-
-    # If there are nontrivial sublattices with vacancies in them,
-    # generate a set of points where their fraction is zero and renormalize
-    for idx, sublattice in enumerate(phase_constituents):
-        if 'VA' in set(sublattice) and len(sublattice) > 1:
-            var_idx = variables.index(v.SiteFraction(phase_name, idx, 'VA'))
-            addtl_pts = np.copy(points)
-            # set vacancy fraction to log-spaced between 1e-10 and 1e-6
-            addtl_pts[:, var_idx] = np.power(10.0, -10.0 * (1.0 - addtl_pts[:, var_idx]))
-            # renormalize site fractions
-            cur_idx = 0
-            for ctx in sublattice_dof:
-                end_idx = cur_idx + ctx
-                addtl_pts[:, cur_idx:end_idx] /= \
-                    addtl_pts[:, cur_idx:end_idx].sum(axis=1)[:, None]
-                cur_idx = end_idx
-            # add to points matrix
-            points = np.concatenate((points, addtl_pts), axis=0)
     # Filter out nan's that may have slipped in if we sampled too high a vacancy concentration
     # Issues with this appear to be platform-dependent
     points = points[~np.isnan(points).any(axis=-1)]
@@ -223,7 +117,7 @@ def _sample_phase_constitution(phase_name, phase_constituents, sublattice_dof, c
     return points
 
 
-def _compute_phase_values(phase_obj, components, variables, statevar_dict,
+def _compute_phase_values(components, statevar_dict,
                           points, phase_record, output, maximum_internal_dof, broadcast=True, fake_points=False,
                           largest_energy=None):
     """
@@ -231,12 +125,8 @@ def _compute_phase_values(phase_obj, components, variables, statevar_dict,
 
     Parameters
     ----------
-    phase_obj : Phase
-        Phase object from a thermodynamic database.
     components : list
         Names of components to consider in the calculation.
-    variables : list
-        Names of variables in the phase's internal degrees of freedom.
     statevar_dict : OrderedDict {str -> float or sequence}
         Mapping of state variables to desired values. This will broadcast if necessary.
     points : ndarray
@@ -282,55 +172,59 @@ def _compute_phase_values(phase_obj, components, variables, statevar_dict,
                                  'broadcast=False.')
             statevars_.append(statevar)
         statevars = statevars_
+    pure_elements = [list(x.constituents.keys()) for x in components]
+    pure_elements = sorted(set([el.upper() for constituents in pure_elements for el in constituents]))
+    pure_elements = [x for x in pure_elements if x != 'VA']
     # func may only have support for vectorization along a single axis (no broadcasting)
     # we need to force broadcasting and flatten the result before calling
-    bc_statevars = [np.ascontiguousarray(broadcast_to(x, points.shape[:-1]).reshape(-1)) for x in statevars]
-    pts = points.reshape(-1, points.shape[-1]).T
-    dof = np.ascontiguousarray(np.concatenate((bc_statevars, pts), axis=0).T)
-    phase_output = np.ascontiguousarray(np.zeros(dof.shape[0]))
-    phase_record.obj(phase_output, dof)
+    bc_statevars = np.ascontiguousarray([broadcast_to(x, points.shape[:-1]).reshape(-1) for x in statevars])
+    pts = points.reshape(-1, points.shape[-1])
+    dof = np.ascontiguousarray(np.concatenate((bc_statevars.T, pts), axis=1))
+    phase_output = np.zeros(dof.shape[0], order='C')
+    phase_compositions = np.zeros((dof.shape[0], len(pure_elements)), order='F')
+    phase_record.obj_2d(phase_output, dof)
+    for el_idx in range(len(pure_elements)):
+        phase_record.mass_obj_2d(phase_compositions[:, el_idx], dof, el_idx)
+
+    max_tieline_vertices = len(pure_elements)
     if isinstance(phase_output, (float, int)):
         phase_output = broadcast_to(phase_output, points.shape[:-1])
+    if isinstance(phase_compositions, (float, int)):
+        phase_compositions = broadcast_to(phase_output, points.shape[:-1] + (len(pure_elements),))
     phase_output = np.asarray(phase_output, dtype=np.float)
     phase_output.shape = points.shape[:-1]
+    phase_compositions = np.asarray(phase_compositions, dtype=np.float)
+    phase_compositions.shape = points.shape[:-1] + (len(pure_elements),)
     if fake_points:
-        phase_output = np.concatenate((broadcast_to(largest_energy, points.shape[:-2] + (len(components),)), phase_output), axis=-1)
-        phase_names = np.concatenate((broadcast_to('_FAKE_', points.shape[:-2] + (len(components),)),
-                                      np.full(points.shape[:-1], phase_obj.name, dtype='U' + str(len(phase_obj.name)))), axis=-1)
+        phase_output = np.concatenate((broadcast_to(largest_energy, points.shape[:-2] + (max_tieline_vertices,)), phase_output), axis=-1)
+        phase_names = np.concatenate((broadcast_to('_FAKE_', points.shape[:-2] + (max_tieline_vertices,)),
+                                      np.full(points.shape[:-1], phase_record.phase_name, dtype='U' + str(len(phase_record.phase_name)))), axis=-1)
     else:
-        phase_names = np.full(points.shape[:-1], phase_obj.name, dtype='U'+str(len(phase_obj.name)))
-
-    # Map the internal degrees of freedom to global coordinates
-    # Normalize site ratios by the sum of site ratios times a factor
-    # related to the site fraction of vacancies
-    site_ratio_normalization = np.zeros(points.shape[:-1])
-    for idx, sublattice in enumerate(phase_obj.constituents):
-        vacancy_column = np.ones(points.shape[:-1])
-        if 'VA' in set(sublattice):
-            var_idx = variables.index(v.SiteFraction(phase_obj.name, idx, 'VA'))
-            vacancy_column -= points[..., :, var_idx]
-        site_ratio_normalization += phase_obj.sublattices[idx] * vacancy_column
-
-    phase_compositions = np.empty(points.shape[:-1] + (len(components),))
-    for col, comp in enumerate(components):
-        avector = [float(vxx.species == comp) * \
-            phase_obj.sublattices[vxx.sublattice_index] for vxx in variables]
-        phase_compositions[..., :, col] = np.divide(np.dot(points[..., :, :], avector),
-                                               site_ratio_normalization)
+        phase_names = np.full(points.shape[:-1], phase_record.phase_name, dtype='U'+str(len(phase_record.phase_name)))
     if fake_points:
-        phase_compositions = np.concatenate((np.broadcast_to(np.eye(len(components)), points.shape[:-2] + (len(components), len(components))), phase_compositions), axis=-2)
+        phase_compositions = np.concatenate((np.broadcast_to(np.eye(len(pure_elements)), points.shape[:-2] + (max_tieline_vertices, len(pure_elements))), phase_compositions), axis=-2)
 
-    coordinate_dict = {'component': components}
+    coordinate_dict = {'component': pure_elements}
     # Resize 'points' so it has the same number of columns as the maximum
     # number of internal degrees of freedom of any phase in the calculation.
     # We do this so that everything is aligned for concat.
     # Waste of memory? Yes, but the alternatives are unclear.
+    # In each case, first check if we need to do this...
+    # It can be expensive for many points (~14s for 500M points)
     if fake_points:
-        expanded_points = np.full(points.shape[:-2] + (len(components)+points.shape[-2], maximum_internal_dof), np.nan)
-        expanded_points[..., len(components):, :points.shape[-1]] = points
+        desired_shape = points.shape[:-2] + (max_tieline_vertices + points.shape[-2], maximum_internal_dof)
+        expanded_points = np.full(desired_shape, np.nan)
+        expanded_points[..., len(pure_elements):, :points.shape[-1]] = points
     else:
-        expanded_points = np.full(points.shape[:-1] + (maximum_internal_dof,), np.nan)
-        expanded_points[..., :points.shape[-1]] = points
+        desired_shape = points.shape[:-1] + (maximum_internal_dof,)
+        if points.shape == desired_shape:
+            expanded_points = points
+        else:
+            # TODO: most optimal solution would be to take pre-extended arrays as an argument and remove this
+            # This still copies the array, but is more efficient than filling
+            # an array with np.nan, then copying the existing points
+            append_nans = np.full(desired_shape[:-1] + (desired_shape[-1] - points.shape[-1],), np.nan)
+            expanded_points = np.append(points, append_nans, axis=-1)
     if broadcast:
         coordinate_dict.update({key: np.atleast_1d(value) for key, value in statevar_dict.items()})
         output_columns = [str(x) for x in statevar_dict.keys()] + ['points']
@@ -345,11 +239,10 @@ def _compute_phase_values(phase_obj, components, variables, statevar_dict,
         # Add state variables as data variables rather than as coordinates
         for sym, vals in zip(statevar_dict.keys(), statevars):
             data_arrays.update({sym: (output_columns, vals)})
-
-    return Dataset(data_arrays, coords=coordinate_dict)
+    return LightDataset(data_arrays, coords=coordinate_dict)
 
 
-def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, broadcast=True, parameters=None, **kwargs):
+def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, broadcast=True, parameters=None, to_xarray=True, **kwargs):
     """
     Sample the property surface of 'output' containing the specified
     components and phases. Model parameters are taken from 'dbf' and any
@@ -403,102 +296,62 @@ def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, bro
     # there may be keyword arguments that aren't state variables
     pdens_dict = unpack_kwarg(kwargs.pop('pdens', 2000), default_arg=2000)
     points_dict = unpack_kwarg(kwargs.pop('points', None), default_arg=None)
-    model_dict = unpack_kwarg(kwargs.pop('model', FallbackModel), default_arg=FallbackModel)
-    callable_dict = unpack_kwarg(kwargs.pop('callables', None), default_arg=None)
+    callables = kwargs.pop('callables', {})
     sampler_dict = unpack_kwarg(kwargs.pop('sampler', None), default_arg=None)
     fixedgrid_dict = unpack_kwarg(kwargs.pop('grid_points', True), default_arg=True)
     parameters = parameters or dict()
     if isinstance(parameters, dict):
         parameters = OrderedDict(sorted(parameters.items(), key=str))
-    param_symbols = tuple(parameters.keys())
-    param_values = np.atleast_1d(np.array(list(parameters.values()), dtype=np.float))
     if isinstance(phases, str):
         phases = [phases]
-    if isinstance(comps, str):
+    if isinstance(comps, (str, v.Species)):
         comps = [comps]
+    comps = sorted(unpack_components(dbf, comps))
     if points_dict is None and broadcast is False:
         raise ValueError('The \'points\' keyword argument must be specified if broadcast=False is also given.')
-    components = [x for x in sorted(comps) if not x.startswith('VA')]
+    nonvacant_components = [x for x in sorted(comps) if x.number_of_atoms > 0]
 
-    # Convert keyword strings to proper state variable objects
-    # If we don't do this, sympy will get confused during substitution
-    statevar_dict = collections.OrderedDict((v.StateVariable(key), unpack_condition(value)) \
-                                            for (key, value) in sorted(kwargs.items()))
-    # XXX: CompiledModel assumes P, T are the only state variables
-    if statevar_dict.get(v.P, None) is None:
-        statevar_dict[v.P] = 101325
-    if statevar_dict.get(v.T, None) is None:
-        statevar_dict[v.T] = 300
-    str_statevar_dict = collections.OrderedDict((str(key), unpack_condition(value)) \
-                                                for (key, value) in statevar_dict.items())
     all_phase_data = []
-    comp_sets = {}
-    largest_energy = 1e30
-    maximum_internal_dof = 0
+    largest_energy = 1e10
 
     # Consider only the active phases
-    active_phases = dict((name.upper(), dbf.phases[name.upper()]) \
-        for name in unpack_phases(phases))
+    list_of_possible_phases = filter_phases(dbf, comps)
+    if len(list_of_possible_phases) == 0:
+        raise ConditionError('There are no phases in the Database that can be active with components {0}'.format(comps))
+    active_phases = {name: dbf.phases[name] for name in filter_phases(dbf, comps, phases)}
+    if len(active_phases) == 0:
+        raise ConditionError('None of the passed phases ({0}) are active. List of possible phases: {1}.'.format(phases, list_of_possible_phases))
 
-    for phase_name, phase_obj in sorted(active_phases.items()):
-        # Build the symbolic representation of the energy
-        mod = model_dict[phase_name]
-        # if this is an object type, we need to construct it
-        if isinstance(mod, type):
-            try:
-                model_dict[phase_name] = mod = mod(dbf, comps, phase_name, parameters=parameters)
-            except DofError:
-                # we can't build the specified phase because the
-                # specified components aren't found in every sublattice
-                # we'll just skip it
-                warnings.warn("""Suspending specified phase {} due to
-                some sublattices containing only unspecified components""".format(phase_name))
-                continue
-        if points_dict[phase_name] is None:
-            maximum_internal_dof = max(maximum_internal_dof, sum(len(x) for x in mod.constituents))
-        else:
-            maximum_internal_dof = max(maximum_internal_dof, np.asarray(points_dict[phase_name]).shape[-1])
+    models = instantiate_models(dbf, comps, list(active_phases.keys()), model=kwargs.pop('model', None), parameters=parameters)
 
+    if isinstance(output, (list, tuple, set)):
+        raise NotImplementedError('Only one property can be specified in calculate() at a time')
+    output = output if output is not None else 'GM'
+
+    # Implicitly add 'N' state variable as a string to keyword arguements if it's not passed
+    if kwargs.get('N') is None:
+        kwargs['N'] = 1
+    if np.any(np.array(kwargs['N']) != 1):
+        raise ConditionError('N!=1 is not yet supported, got N={}'.format(kwargs['N']))
+
+    # TODO: conditions dict of StateVariable instances should become part of the calculate API
+    statevar_strings = [sv for sv in kwargs.keys() if getattr(v, sv) is not None]
+    # If we don't do this, sympy will get confused during substitution
+    statevar_dict = dict((v.StateVariable(key), unpack_condition(value)) for key, value in kwargs.items() if key in statevar_strings)
+    # Sort after default state variable check to fix gh-116
+    statevar_dict = collections.OrderedDict(sorted(statevar_dict.items(), key=lambda x: str(x[0])))
+    phase_records = build_phase_records(dbf, comps, active_phases, statevar_dict,
+                                   models=models, parameters=parameters,
+                                   output=output, callables=callables, build_gradients=False, build_hessians=False,
+                                   verbose=kwargs.pop('verbose', False))
+    str_statevar_dict = collections.OrderedDict((str(key), unpack_condition(value)) \
+                                                for (key, value) in statevar_dict.items())
+    maximum_internal_dof = max(len(models[phase_name].site_fractions) for phase_name in active_phases)
     for phase_name, phase_obj in sorted(active_phases.items()):
-        try:
-            mod = model_dict[phase_name]
-        except KeyError:
-            continue
-        # this is a phase model we couldn't construct for whatever reason; skip it
-        if isinstance(mod, type):
-            continue
-        if (not isinstance(mod, CompiledModel)) or (output != 'GM'):
-            if isinstance(mod, CompiledModel):
-                mod = Model(dbf, comps, phase_name, parameters=parameters)
-            # Construct an ordered list of the variables
-            variables, sublattice_dof = generate_dof(phase_obj, mod.components)
-            # Build the "fast" representation of that model
-            if callable_dict[phase_name] is None:
-                try:
-                    out = getattr(mod, output)
-                except AttributeError:
-                    raise AttributeError('Missing Model attribute {0} specified for {1}'
-                                         .format(output, mod.__class__))
-                # As a last resort, treat undefined symbols as zero
-                # But warn the user when we do this
-                # This is consistent with TC's behavior
-                undefs = list(out.atoms(Symbol) - out.atoms(v.StateVariable))
-                for undef in undefs:
-                    out = out.xreplace({undef: float(0)})
-                    warnings.warn('Setting undefined symbol {0} for phase {1} to zero'.format(undef, phase_name))
-                comp_sets[phase_name] = build_functions(out, list(statevar_dict.keys()) + variables,
-                                                        include_obj=True, include_grad=False, include_hess=False,
-                                                        parameters=param_symbols)
-            else:
-                comp_sets[phase_name] = callable_dict[phase_name]
-            phase_record = PhaseRecord_from_cython(comps, list(statevar_dict.keys()) + variables,
-                                        np.array(dbf.phases[phase_name].sublattices, dtype=np.float),
-                                        param_values, comp_sets[phase_name], None, None)
-        else:
-            variables = sorted(set(mod.variables) - {v.T, v.P}, key=str)
-            sublattice_dof = mod.sublattice_dof
-            phase_record = PhaseRecord_from_compiledmodel(mod, param_values)
+        mod = models[phase_name]
+        phase_record = phase_records[phase_name]
         points = points_dict[phase_name]
+        variables, sublattice_dof = generate_dof(phase_obj, mod.components)
         if points is None:
             points = _sample_phase_constitution(phase_name, phase_obj.constituents, sublattice_dof, comps,
                                                 tuple(variables), sampler_dict[phase_name] or point_sample,
@@ -506,7 +359,7 @@ def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, bro
         points = np.atleast_2d(points)
 
         fp = fake_points and (phase_name == sorted(active_phases.keys())[0])
-        phase_ds = _compute_phase_values(phase_obj, components, variables, str_statevar_dict,
+        phase_ds = _compute_phase_values(nonvacant_components, str_statevar_dict,
                                          points, phase_record, output,
                                          maximum_internal_dof, broadcast=broadcast,
                                          largest_energy=float(largest_energy), fake_points=fp)
@@ -514,9 +367,22 @@ def calculate(dbf, comps, phases, mode=None, output='GM', fake_points=False, bro
 
     # speedup for single-phase case (found by profiling)
     if len(all_phase_data) > 1:
-        final_ds = _fast_concat(all_phase_data, dim='points')
-        final_ds['points'].values = np.arange(len(final_ds['points']))
-        final_ds.coords['points'].values = np.arange(len(final_ds['points']))
+        concatenated_coords = all_phase_data[0].coords
+
+        data_vars = all_phase_data[0].data_vars
+        concatenated_data_vars = {}
+        for var in data_vars.keys():
+            data_coords = data_vars[var][0]
+            points_idx = data_coords.index('points')  # concatenation axis
+            arrs = []
+            for phase_data in all_phase_data:
+                arrs.append(getattr(phase_data, var))
+            concat_data = np.concatenate(arrs, axis=points_idx)
+            concatenated_data_vars[var] = (data_coords, concat_data)
+        final_ds = LightDataset(data_vars=concatenated_data_vars, coords=concatenated_coords)
     else:
         final_ds = all_phase_data[0]
-    return final_ds
+    if to_xarray:
+        return final_ds.get_dataset()
+    else:
+        return final_ds
