@@ -11,11 +11,14 @@ import re
 from sympy import sympify, And, Or, Not, Intersection, Union, EmptySet, Interval, Piecewise
 from sympy import Symbol, GreaterThan, StrictGreaterThan, LessThan, StrictLessThan, Complement, S
 from sympy import Mul, Pow, Rational
+from sympy.abc import _clash
 from sympy.printing.str import StrPrinter
 from sympy.core.mul import _keep_coeff
 from sympy.printing.precedence import precedence
 from pycalphad import Database
 from pycalphad.io.database import DatabaseExportError
+from pycalphad.io.grammar import float_number, chemical_formula
+from pycalphad.variables import Species
 import pycalphad.variables as v
 from pycalphad.io.tdb_keywords import expand_keyword, TDB_PARAM_TYPES
 from collections import defaultdict, namedtuple
@@ -30,9 +33,21 @@ import warnings
 import hashlib
 from copy import deepcopy
 
-_AST_WHITELIST = [ast.Add, ast.BinOp, ast.Call, ast.Div, ast.Expression,
-                  ast.Load, ast.Mult, ast.Name, ast.Num, ast.Pow, ast.Sub,
-                  ast.UAdd, ast.UnaryOp, ast.USub]
+# ast.Num is deprecated in Python 3.8 in favor of as ast.Constant
+# Both are whitelisted for compatability across versions
+_AST_WHITELIST = [ast.Add, ast.BinOp, ast.Call, ast.Constant, ast.Div,
+                  ast.Expression, ast.Load, ast.Mult, ast.Name, ast.Num,
+                  ast.Pow, ast.Sub, ast.UAdd, ast.UnaryOp, ast.USub]
+
+# Avoid symbol names clashing with objects in sympy (gh-233)
+clashing_namespace = {}
+clashing_namespace.update(_clash)
+clashing_namespace['CC'] = Symbol('CC')
+clashing_namespace['FF'] = Symbol('FF')
+clashing_namespace['T'] = v.T
+clashing_namespace['P'] = v.P
+clashing_namespace['R'] = v.R
+
 
 def _sympify_string(math_string):
     "Convert math string into SymPy object."
@@ -47,12 +62,7 @@ def _sympify_string(math_string):
     expr_string = \
         re.sub(r'(?<!\w)EXP(?!\w)', 'exp', expr_string,
                flags=re.IGNORECASE)
-    # Convert raw variables into StateVariable objects
-    variable_fixes = {
-        Symbol('T'): v.T,
-        Symbol('P'): v.P,
-        Symbol('R'): v.R
-    }
+
     # sympify uses eval, so we need to sanitize the input
     nodes = ast.parse(expr_string)
     nodes = ast.Expression(nodes.body[0].value)
@@ -61,7 +71,8 @@ def _sympify_string(math_string):
         if type(node) not in _AST_WHITELIST: #pylint: disable=W1504
             raise ValueError('Expression from TDB file not in whitelist: '
                              '{}'.format(expr_string))
-    return sympify(expr_string).xreplace(variable_fixes)
+
+    return sympify(expr_string, locals=clashing_namespace)
 
 def _parse_action(func):
     """
@@ -82,7 +93,10 @@ def _parse_action(func):
     Source: Florian Brucker on StackOverflow
     http://stackoverflow.com/questions/10177276/pyparsing-setparseaction-function-is-getting-no-arguments
     """
-    num_args = len(inspect.getargspec(func).args)
+    func_items = inspect.signature(func).parameters.items()
+    func_args = [name for name, param in func_items
+                 if param.kind == param.POSITIONAL_OR_KEYWORD]
+    num_args = len(func_args)
     if num_args > 3:
         raise ValueError('Input function must take at most 3 parameters.')
 
@@ -138,7 +152,7 @@ def _make_piecewise_ast(toks):
             )
         cur_tok = cur_tok + 2
     expr_cond_pairs.append((0, True))
-    return Piecewise(*expr_cond_pairs)
+    return Piecewise(*expr_cond_pairs, evaluate=False)
 
 class TCCommand(CaselessKeyword): #pylint: disable=R0903
     """
@@ -178,26 +192,25 @@ def _tdb_grammar(): #pylint: disable=R0914
     Convenience function for getting the pyparsing grammar of a TDB file.
     """
     int_number = Word(nums).setParseAction(lambda t: [int(t[0])])
-    # matching float w/ regex is ugly but is recommended by pyparsing
-    float_number = Regex(r'[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?') \
-        .setParseAction(lambda t: [float(t[0])])
     # symbol name, e.g., phase name, function name
     symbol_name = Word(alphanums+'_:', min=1)
-    ref_phase_name = symbol_name = Word(alphanums+'_:()', min=1)
+    ref_phase_name = symbol_name = Word(alphanums+'_-:()/', min=1)
     # species name, e.g., CO2, AL, FE3+
-    species_name = Word(alphanums+'+-*', min=1) + Optional(Suppress('%'))
+    species_name = Word(alphanums+'+-*/_.', min=1) + Optional(Suppress('%'))
     # constituent arrays are colon-delimited
     # each subarray can be comma- or space-delimited
     constituent_array = Group(delimitedList(Group(OneOrMore(Optional(Suppress(',')) + species_name)), ':'))
     param_types = MatchFirst([TCCommand(param_type) for param_type in TDB_PARAM_TYPES])
     # Let sympy do heavy arithmetic / algebra parsing for us
     # a convenience function will handle the piecewise details
-    func_expr = Optional(float_number) + OneOrMore(SkipTo(';') \
+    func_expr = (float_number | ZeroOrMore(',').setParseAction(lambda t: 0.01)) + OneOrMore(SkipTo(';') \
         + Suppress(';') + ZeroOrMore(Suppress(',')) + Optional(float_number) + \
         Suppress(Word('YNyn', exact=1) | White()))
     # ELEMENT
-    cmd_element = TCCommand('ELEMENT') + Word(alphas+'/-', min=1, max=2) + Optional(Suppress(ref_phase_name)) + \
-        Optional(Suppress(OneOrMore(float_number))) + LineEnd()
+    cmd_element = TCCommand('ELEMENT') + Word(alphas+'/-', min=1, max=2) + ref_phase_name + \
+        float_number + float_number + float_number + LineEnd()
+    # SPECIES
+    cmd_species = TCCommand('SPECIES') + species_name + chemical_formula + LineEnd()
     # TYPE_DEFINITION
     cmd_typedef = TCCommand('TYPE_DEFINITION') + \
         Suppress(White()) + CharsNotIn(' !', exact=1) + SkipTo(LineEnd())
@@ -210,8 +223,18 @@ def _tdb_grammar(): #pylint: disable=R0914
     cmd_defsysdef = TCCommand('DEFINE_SYSTEM_DEFAULT') + SkipTo(LineEnd())
     # DEFAULT_COMMAND
     cmd_defcmd = TCCommand('DEFAULT_COMMAND') + SkipTo(LineEnd())
+    # DATABASE_INFO
+    cmd_database_info = TCCommand('DATABASE_INFO') + SkipTo(LineEnd())
+    # VERSION_DATE
+    cmd_version_date = TCCommand('VERSION_DATE') + SkipTo(LineEnd())
+    # REFERENCE_FILE
+    cmd_reference_file = TCCommand('REFERENCE_FILE') + SkipTo(LineEnd())
+    # ADD_REFERENCES
+    cmd_add_ref = TCCommand('ADD_REFERENCES') + SkipTo(LineEnd())
     # LIST_OF_REFERENCES
     cmd_lor = TCCommand('LIST_OF_REFERENCES') + SkipTo(LineEnd())
+    # TEMPERATURE_LIMITS
+    cmd_templim = TCCommand('TEMPERATURE_LIMITS') + SkipTo(LineEnd())
     # PHASE
     cmd_phase = TCCommand('PHASE') + symbol_name + \
         Suppress(White()) + CharsNotIn(' !', min=1) + Suppress(White()) + \
@@ -229,12 +252,18 @@ def _tdb_grammar(): #pylint: disable=R0914
         Suppress(')') + func_expr.setParseAction(_make_piecewise_ast)
     # Now combine the grammar together
     all_commands = cmd_element | \
+                    cmd_species | \
                     cmd_typedef | \
                     cmd_function | \
                     cmd_ass_sys | \
                     cmd_defsysdef | \
                     cmd_defcmd | \
+                    cmd_database_info | \
+                    cmd_version_date | \
+                    cmd_reference_file | \
+                    cmd_add_ref | \
                     cmd_lor | \
+                    cmd_templim | \
                     cmd_phase | \
                     cmd_constituent | \
                     cmd_parameter
@@ -242,37 +271,67 @@ def _tdb_grammar(): #pylint: disable=R0914
 
 def _process_typedef(targetdb, typechar, line):
     """
-    Process the TYPE_DEFINITION command.
+    Process a TYPE_DEFINITION command.
+
+    Assumes all phases are entered into the database already and that the
+    database defines _typechar_map, which defines a map of typechar to the
+    phases that use it. Any phases that in the typechar dict for this will have
+    the model_hints updated based on this type definition, regardless of which
+    phase names may be defined in this TYPE_DEF line.
+
     """
+    matching_phases = targetdb._typechar_map[typechar]
+    del targetdb._typechar_map[typechar]
     # GES A_P_D BCC_A2 MAGNETIC  -1    0.4
     tokens = line.replace(',', '').split()
     if len(tokens) < 4:
         return
     keyword = expand_keyword(['DISORDERED_PART', 'MAGNETIC'], tokens[3].upper())[0]
     if len(keyword) == 0:
-        raise ValueError('Unknown keyword: {}'.format(tokens[3]))
+        raise ValueError('Unknown type definition keyword: {}'.format(tokens[3]))
+    if len(matching_phases) == 0:
+        warnings.warn(f"The type definition character `{typechar}` in `TYPE_DEFINITION {typechar} {line}` is not used by any phase.")
     if keyword == 'MAGNETIC':
-        # magnetic model (IHJ model assumed by default)
-        targetdb.tdbtypedefs[typechar] = {
-            'ihj_magnetic':[float(tokens[4]), float(tokens[5])]
+        # Magnetic model, both IHJ and Xiong models use these model hints when
+        # constructing Model instances, despite being prefixed `ihj_magnetic_`
+        model_hints = {
+            'ihj_magnetic_afm_factor': float(tokens[4]),
+            'ihj_magnetic_structure_factor': float(tokens[5])
         }
+        for phase_name in matching_phases:
+            targetdb.phases[phase_name].model_hints.update(model_hints)
+
     # GES A_P_D L12_FCC DIS_PART FCC_A1
     if keyword == 'DISORDERED_PART':
-        # order-disorder model
-        targetdb.tdbtypedefs[typechar] = {
-            'disordered_phase': tokens[4].upper(),
-            'ordered_phase': tokens[2].upper()
+        # order-disorder model: since we need to add model_hints to both the
+        # ordered and disorderd phase, we special case to update the phase
+        # names defined by the TYPE_DEF, rather than the updating the phases
+        # with matching typechars.
+        ordered_phase = tokens[2].upper()
+        disordered_phase = tokens[4].upper()
+        hint = {
+            'ordered_phase': ordered_phase,
+            'disordered_phase': disordered_phase,
         }
-        if tokens[2].upper() in targetdb.phases:
-            # Since TDB files do not enforce any kind of ordering
-            # on the specification of ordered and disordered phases,
-            # we need to handle the case of when either phase is specified
-            # first. In this case, we imagine the ordered phase is
-            # specified first. If the disordered phase is specified
-            # first, we will have to catch it in _process_phase().
-            targetdb.phases[tokens[2].upper()].model_hints.update(
-                targetdb.tdbtypedefs[typechar]
-            )
+        if ordered_phase in targetdb.phases:
+            targetdb.phases[ordered_phase].model_hints.update(hint)
+        else:
+            raise ValueError(f"The {ordered_phase} phase is not in the database, but is defined by: `TYPE_DEFINTION {typechar} {line}`")
+        if disordered_phase in targetdb.phases:
+            targetdb.phases[disordered_phase].model_hints.update(hint)
+        else:
+            raise ValueError(f"The {disordered_phase} phase is not in the database, but is defined by: `TYPE_DEFINTION {typechar} {line}`")
+
+
+phase_options = {'ionic_liquid_2SL': 'Y',
+                 'symmetry_FCC_4SL': 'F',
+                 'symmetry_BCC_4SL': 'B',
+                 'liquid': 'L',
+                 'gas': 'G',
+                 'aqueous': 'A',
+                 'charged_phase': 'I'}
+inv_phase_options = dict([reversed(i) for i in phase_options.items()])
+
 
 def _process_phase(targetdb, name, typedefs, subls):
     """
@@ -280,27 +339,19 @@ def _process_phase(targetdb, name, typedefs, subls):
     """
     splitname = name.split(':')
     phase_name = splitname[0].upper()
-    options = None
+    options = ''
     if len(splitname) > 1:
         options = splitname[1]
     targetdb.add_structure_entry(phase_name, phase_name)
     model_hints = {}
-    for typedef in list(typedefs):
-        if typedef in targetdb.tdbtypedefs.keys():
-            if 'ihj_magnetic' in targetdb.tdbtypedefs[typedef].keys():
-                model_hints['ihj_magnetic_afm_factor'] = \
-                    targetdb.tdbtypedefs[typedef]['ihj_magnetic'][0]
-                model_hints['ihj_magnetic_structure_factor'] = \
-                    targetdb.tdbtypedefs[typedef]['ihj_magnetic'][1]
-            if 'ordered_phase' in targetdb.tdbtypedefs[typedef].keys():
-                model_hints['ordered_phase'] = \
-                    targetdb.tdbtypedefs[typedef]['ordered_phase']
-                model_hints['disordered_phase'] = \
-                    targetdb.tdbtypedefs[typedef]['disordered_phase']
-                if model_hints['disordered_phase'] in targetdb.phases:
-                    targetdb.phases[model_hints['disordered_phase']]\
-                        .model_hints.update({'ordered_phase': model_hints['ordered_phase'],
-                                             'disordered_phase': model_hints['disordered_phase']})
+    for option in inv_phase_options.keys():
+        if option in options:
+            model_hints[inv_phase_options[option]] = True
+
+    for typedef_char in list(typedefs):
+        targetdb._typechar_map[typedef_char].append(phase_name)
+
+    # Model hints are updated later based on the type definitions
     targetdb.add_phase(phase_name, model_hints, subls)
 
 def _process_parameter(targetdb, param_type, phase_name, diffusing_species,
@@ -312,7 +363,7 @@ def _process_parameter(targetdb, param_type, phase_name, diffusing_species,
     targetdb.add_parameter(param_type, phase_name.upper(),
                            [[c.upper() for c in sorted(lx)]
                             for lx in constituent_array.asList()],
-                           param_order, param, ref, diffusing_species)
+                           param_order, param, ref, diffusing_species, force_insert=False)
 
 def _unimplemented(*args, **kwargs): #pylint: disable=W0613
     """
@@ -320,19 +371,39 @@ def _unimplemented(*args, **kwargs): #pylint: disable=W0613
     """
     pass
 
+def _process_species(db, sp_name, sp_comp, charge=0, *args):
+    """Add a species to the Database. If charge not specified, the Species will be neutral."""
+    # process the species composition list of [element1, ratio1, element2, ratio2, ..., elementN, ratioN]
+    constituents = {sp_comp[i]: sp_comp[i+1] for i in range(0, len(sp_comp), 2)}
+    db.species.add(Species(sp_name, constituents, charge=charge))
+
+def _process_reference_state(db, el, refphase, mass, H298, S298):
+    db.refstates[el] = {
+        'phase': refphase,
+        'mass': mass,
+        'H298': H298,
+        'S298': S298,
+    }
+
 def _setitem_raise_duplicates(dictionary, key, value):
     if key in dictionary:
         raise ValueError("TDB contains duplicate FUNCTION {}".format(key))
     dictionary[key] = value
 
 _TDB_PROCESSOR = {
-    'ELEMENT': lambda db, el: db.elements.add(el),
-    'TYPE_DEFINITION': _process_typedef,
+    'ELEMENT': lambda db, el, ref_phase, mass, h, s: (db.elements.add(el), _process_reference_state(db, el, ref_phase, mass, h, s), _process_species(db, el, [el, 1], 0)),
+    'SPECIES': _process_species,
+    'TYPE_DEFINITION': lambda db, typechar, line: db._typedefs_queue.append((typechar, line)),
     'FUNCTION': lambda db, name, sym: _setitem_raise_duplicates(db.symbols, name, sym),
     'DEFINE_SYSTEM_DEFAULT': _unimplemented,
     'ASSESSED_SYSTEMS': _unimplemented,
     'DEFAULT_COMMAND': _unimplemented,
+    'DATABASE_INFO': _unimplemented,
+    'VERSION_DATE': _unimplemented,
+    'REFERENCE_FILE': _unimplemented,
+    'ADD_REFERENCES': _unimplemented,
     'LIST_OF_REFERENCES': _unimplemented,
+    'TEMPERATURE_LIMITS': _unimplemented,
     'PHASE': _process_phase,
     'CONSTITUENT': \
         lambda db, name, c: db.add_phase_constituents(
@@ -342,11 +413,11 @@ _TDB_PROCESSOR = {
 
 def to_interval(relational):
     if isinstance(relational, And):
-        return Intersection([to_interval(i) for i in relational.args])
+        return Intersection(*[to_interval(i) for i in relational.args])
     elif isinstance(relational, Or):
-        return Union([to_interval(i) for i in relational.args])
+        return Union(*[to_interval(i) for i in relational.args])
     elif isinstance(relational, Not):
-        return Complement([to_interval(i) for i in relational.args])
+        return Complement(*[to_interval(i) for i in relational.args])
     if relational == S.true:
         return Interval(S.NegativeInfinity, S.Infinity, left_open=True, right_open=True)
 
@@ -392,9 +463,9 @@ class TCPrinter(StrPrinter):
         # Need to verify that each cond's highlim equals the next cond's lowlim
         # to_interval() is used instead of sympy.Relational.as_set() for performance reasons
         intervals = [to_interval(i.cond) for i in filtered_args]
-        if (len(intervals) > 1) and Intersection(intervals) != EmptySet():
+        if (len(intervals) > 1) and not Intersection(*intervals) is EmptySet:
             raise ValueError('Overlapping intervals cannot be represented: {}'.format(intervals))
-        if not isinstance(Union(intervals), Interval):
+        if not isinstance(Union(*intervals), Interval):
             raise ValueError('Piecewise intervals must be continuous')
         if not all([arg.cond.free_symbols == {v.T} for arg in filtered_args]):
             raise ValueError('Only temperature-dependent piecewise conditions are supported')
@@ -585,7 +656,7 @@ def _apply_new_symbol_names(dbf, symbol_name_map):
     dbf.symbols = {name: S(expr).xreplace({Symbol(s): Symbol(v) for s, v in symbol_name_map.items()}) for name, expr in dbf.symbols.items()}
     # finally propagate through to the parameters
     for p in dbf._parameters.all():
-        dbf._parameters.update({'parameter': S(p['parameter']).xreplace({Symbol(s): Symbol(v) for s, v in symbol_name_map.items()})}, eids=[p.eid])
+        dbf._parameters.update({'parameter': S(p['parameter']).xreplace({Symbol(s): Symbol(v) for s, v in symbol_name_map.items()})}, doc_ids=[p.doc_id])
 
 
 def write_tdb(dbf, fd, groupby='subsystem', if_incompatible='warn'):
@@ -656,11 +727,27 @@ def write_tdb(dbf, fd, groupby='subsystem', if_incompatible='warn'):
     output += "$ Generated by {} (pycalphad {})\n".format(getpass.getuser(), __version__)
     output += ("$" * maxlen) + "\n\n"
     for element in sorted(dbf.elements):
-        output += "ELEMENT {0} BLANK 0 0 0 !\n".format(element.upper())
+        ref = dbf.refstates.get(element, {})
+        refphase = ref.get('phase', 'BLANK')
+        mass = ref.get('mass', 0.0)
+        H298 = ref.get('H298', 0.0)
+        S298 = ref.get('S298', 0.0)
+        output += "ELEMENT {0} {1} {2} {3} {4} !\n".format(element.upper(), refphase, mass, H298, S298)
     if len(dbf.elements) > 0:
         output += "\n"
-    for species in sorted(dbf.species):
-        output += "SPECIES {0} !\n".format(species.upper())
+    for species in sorted(dbf.species, key=lambda s: s.name):
+        if species.name not in dbf.elements:
+            # construct the charge part of the specie
+            if species.charge != 0:
+                if species.charge >0:
+                    charge_sign = '+'
+                else:
+                    charge_sign = ''
+                charge = '/{}{}'.format(charge_sign, species.charge)
+            else:
+                charge = ''
+            species_constituents = ''.join(['{}{}'.format(el, val) for el, val in sorted(species.constituents.items(), key=lambda t: t[0])])
+            output += "SPECIES {0} {1}{2} !\n".format(species.name.upper(), species_constituents, charge)
     if len(dbf.species) > 0:
         output += "\n"
     # Write FUNCTION block
@@ -686,6 +773,10 @@ def write_tdb(dbf, fd, groupby='subsystem', if_incompatible='warn'):
     typedefs = defaultdict(lambda: ["%"])
     for name, phase_obj in sorted(dbf.phases.items()):
         model_hints = phase_obj.model_hints.copy()
+        possible_options = set(phase_options.keys()).intersection(model_hints)
+        # Phase options are handled later
+        for option in possible_options:
+            del model_hints[option]
         if ('ordered_phase' in model_hints.keys()) and (model_hints['ordered_phase'] == name):
             new_char = typedef_chars.pop()
             typedefs[name].append(new_char)
@@ -712,11 +803,19 @@ def write_tdb(dbf, fd, groupby='subsystem', if_incompatible='warn'):
             raise ValueError('Not all model hints are supported: {}'.format(model_hints))
     # Perform a second loop now that all typedefs / model hints are consistent
     for name, phase_obj in sorted(dbf.phases.items()):
-        output += "PHASE {0} {1}  {2} {3} !\n".format(name.upper(), ''.join(typedefs[name]),
+        # model_hints may also contain "phase options", e.g., ionic liquid
+        model_hints = phase_obj.model_hints.copy()
+        name_with_options = str(name.upper())
+        possible_options = set(phase_options.keys()).intersection(model_hints.keys())
+        if len(possible_options) > 0:
+            name_with_options += ':'
+        for option in possible_options:
+            name_with_options += phase_options[option]
+        output += "PHASE {0} {1}  {2} {3} !\n".format(name_with_options, ''.join(typedefs[name]),
                                                       len(phase_obj.sublattices),
                                                       ' '.join([str(i) for i in phase_obj.sublattices]))
-        constituents = ':'.join([','.join(sorted(subl)) for subl in phase_obj.constituents])
-        output += "CONSTITUENT {0} :{1}: !\n".format(name.upper(), constituents)
+        constituents = ':'.join([','.join([spec.name for spec in sorted(subl)]) for subl in phase_obj.constituents])
+        output += "CONSTITUENT {0} :{1}: !\n".format(name_with_options, constituents)
         output += "\n"
 
     # PARAMETERs by subsystem
@@ -728,10 +827,12 @@ def write_tdb(dbf, fd, groupby='subsystem', if_incompatible='warn'):
             components = set()
             for subl in param['constituent_array']:
                 components |= set(subl)
-            if param['diffusing_species'] is not None:
+            if param['diffusing_species'] != Species(None):
                 components |= {param['diffusing_species']}
             # Wildcard operator is not a component
             components -= {'*'}
+            desired_active_pure_elements = [list(x.constituents.keys()) for x in components]
+            components = set([el.upper() for constituents in desired_active_pure_elements for el in constituents])
             # Remove vacancy if it's not the only component (pure vacancy endmember)
             if len(components) > 1:
                 components -= {'VA'}
@@ -749,7 +850,7 @@ def write_tdb(dbf, fd, groupby='subsystem', if_incompatible='warn'):
                                                  param['reference']))
 
     def write_parameter(param_to_write):
-        constituents = ':'.join([','.join(sorted([i.upper() for i in subl]))
+        constituents = ':'.join([','.join(sorted([i.name.upper() for i in subl]))
                          for subl in param_to_write.constituent_array])
         # TODO: Handle references
         paramx = param_to_write.parameter
@@ -760,8 +861,8 @@ def write_tdb(dbf, fd, groupby='subsystem', if_incompatible='warn'):
         exprx = TCPrinter().doprint(paramx).upper()
         if ';' not in exprx:
             exprx += '; N'
-        if param_to_write.diffusing_species is not None:
-            ds = "&" + param_to_write.diffusing_species
+        if param_to_write.diffusing_species != Species(None):
+            ds = "&" + param_to_write.diffusing_species.name
         else:
             ds = ""
         return "PARAMETER {}({}{},{};{}) {} !\n".format(param_to_write.parameter_type.upper(),
@@ -771,8 +872,8 @@ def write_tdb(dbf, fd, groupby='subsystem', if_incompatible='warn'):
                                                         param_to_write.parameter_order,
                                                         exprx)
     if groupby == 'subsystem':
-        for num_elements in range(1, 5):
-            subsystems = list(itertools.combinations(sorted([i.upper() for i in dbf.elements]), num_elements))
+        for num_species in range(1, 5):
+            subsystems = list(itertools.combinations(sorted([i.name.upper() for i in dbf.species]), num_species))
             for subsystem in subsystems:
                 parameters = sorted(param_sorted[subsystem])
                 if len(parameters) > 0:
@@ -784,7 +885,7 @@ def write_tdb(dbf, fd, groupby='subsystem', if_incompatible='warn'):
                     for parameter in parameters:
                         output += write_parameter(parameter)
         # Don't generate combinatorics for multi-component subsystems or we'll run out of memory
-        if len(dbf.elements) > 4:
+        if len(dbf.species) > 4:
             subsystems = [k for k in param_sorted.keys() if len(k) > 4]
             for subsystem in subsystems:
                 parameters = sorted(param_sorted[subsystem])
@@ -834,21 +935,39 @@ def read_tdb(dbf, fd):
     # Now split by the command delimeter
     commands = lines.split('!')
 
-    # Temporary storage while we process type definitions
-    dbf.tdbtypedefs = {}
+    # Temporarily track which typedef characters were used by which phase
+    # before we process the type definitions
+    # Map {typedef character: [phases using that typedef]}
+    dbf._typechar_map = defaultdict(list)
+    dbf._typedefs_queue = []  # queue of type defintion lines to process
+
+    grammar = _tdb_grammar()
 
     for command in commands:
         if len(command) == 0:
             continue
         tokens = None
         try:
-            tokens = _tdb_grammar().parseString(command)
+            tokens = grammar.parseString(command)
             _TDB_PROCESSOR[tokens[0]](dbf, *tokens[1:])
-        except ParseException:
+        except:
             print("Failed while parsing: " + command)
             print("Tokens: " + str(tokens))
             raise
-    del dbf.tdbtypedefs
+
+    # Process type definitions last, updating model_hints for defined phases.
+    for typechar, line in dbf._typedefs_queue:
+        _process_typedef(dbf, typechar, line)
+    # Raise warnings if there are any remaining type characters that one or more
+    # phases expected to be defined
+    for typechar, phases_expecting_typechar in dbf._typechar_map.items():
+        warnings.warn(f"The type definition character `{typechar}` was defined in the following phases: "
+                      f"{phases_expecting_typechar}, but no corresponding TYPE_DEFINITION line was found in the TDB.")
+    del dbf._typechar_map
+    del dbf._typedefs_queue
+
+    dbf.process_parameter_queue()
+
 
 
 Database.register_format("tdb", read=read_tdb, write=write_tdb)
